/* Heuristica de Insercao
 * Este programa encontra uma solucao para o problema do caixeiro viajante utilizando o metodo de insercao.
 * O metodo consiste em, partindo-se de um roteiro inicial com, geralmente, tres cidades e considerando todas as demais
 * cidades ainda nao incluidas no roteiro, seleciona-se para insercao, no roteiro parcial, aquela que atenda a um
 * determinado criterio. Esse procedimento e repetido sucessivamente, com a analise de insercao entre cada par de cidades
 * do roteiro parcial, ate que todas as cidades sejam inseridas na solucao.
 * Alguns dos critrios mais utilizados para a seleo dos vrtices a serem acrescidos ao sub-ciclo so: insero do vrtice
 * mais prximo, insero do vrtice mais distante, insero do vrtice que conduz ao ciclo mais barato (insero mais
 * barata) e insero aleatria. Neste programa foi utilizado o critrio da insero mais barata.
 * Autora: Larissa Alves Petri     #USP: 5117594
 * e-mail: lariss@gmail.com
 * Data: /06/2006
 */


// Bibliotecas utilizadas:

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


// Definicao das constantes:

#define N 6 // Numero de cidades do problema.
#define T 30 // Tamanho do nome do arquivo de entrada.
#define PI 3.141592 // Valor de pi.
#define R 6378.388 
#define INF 32000000 // Infinito.


// Declaracao das funcoes:

void le_arq (float x[N], float y[N], char c[N])
/* Funcao que le o arquivo de entrada, que  contem as coordenadas das cidades. Esse arquivo deve ser .txt.
 * x[N] e y[N]: vetores que receberao as coordenadas.
 * c[N]: nome do arquivo.
 */
{
	int i;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as coordenadas e guarda-as em dois vetores, x e y.
	for (i = 0; i < N; i++)
		fscanf (e, "%*d %f %f\n", &x[i], &y[i]);

	// Fecha o arquivo.
	fclose (e);
}

void le_arq_matriz (int **d, char c[N])
/* Funcao que le o arquivo de entrada, que  contem as distancias entra as cidades. Esse arquivo deve ser .txt.
 * **d: matriz que recebera as distancias.
 * c[N]: nome do arquivo.
 */
{
	int i, j;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as distancias e guarda-as na matriz d.
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < (N - 1); j++)
			fscanf (e, " %d", &d[i][j]);
		fscanf (e, " %d\n", &d[i][j]);
	}

	// Fecha o arquivo.
	fclose (e);
}

void att (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia pseudo-euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo ATT).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, t[N][N];
	float xd, yd, r[N][N];

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			r[i][j] = sqrt ((xd * xd + yd * yd) / 10.0);
			t[i][j] = (int)(r[i][j]);
			if (t[i][j] < r[i][j])
				d[i][j] = t[i][j] + 1;
			else
				d[i][j] = t[i][j];
		}
}

void euc (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo EUC_2D).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j;
	float xd, yd;
	
	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			d[i][j] = (int)(sqrt (xd * xd + yd * yd));
		}
}

void geo (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia geometrica entre as cidades e guarda-a na matriz d (arquivos do tipo GEO).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, deg;
	float min, q1, q2, q3;

	for (i = 0; i < N; i++)
	{
		deg = (int)(x[i]);
		min = x[i] - deg;
		x[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
		deg = (int)(y[i]);
		min = y[i] - deg;
		y[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
	}

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			q1 = cos (y[i] - y[j]);
			q2 = cos (x[i] - x[j]);
			q3 = cos (x[i] + x[j]);
			d[i][j] = (int)(R * acos (0.5 * ((1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0);
		}
}

int verifica (int v[N], int e)
/* Funcao que verifica se o elemento a esta no vetor v.
 * v[N]: vetor solucao inicial.
 * e: elemento que sera testado.
 */
{
	int j = 0;

	// Percorre o vetor, procurando pelo elemento.
	while ((j < N) && (v[j] != e))
		j++;

	// Retorna 1 se o elemento esta no vetor e 0 caso contrario. 
	j = (j != N) ? 1 : 0;

	return j;
}

void insere (int v[N], int e, int p)
/* Funcao que insere o elemento e na posicao p do vetor v.
 * v[N]: vetor solucao.
 * e: elemento a ser inserido.
 * p: posicao do vetor em que o elemento deve ser inserido.
 */
{
	int i, j, k;

	// Inicializacao das variaveis.
    	j = v[p];
    	i = p + 1;

	// Rearranjo dos elementos do vetor, para insercao do novo elemento no meio do vetor.
    	while (i < N)
    	{
    		k = v[i];
        	v[i] = j;
        	j = k;
		// Testa se j eh diferente de zero, ou seja, se ainda ha elementos, nao nulos, no vetor.
	        if (!j)
            		i = N;
        		i++;
    	}
	
    	v[p] = e;
}

void insercao (int v[N], int **d)
/* Funcao que gera uma solucao inicial pela heuristica de insercao.
 * v[N]: vetor que recebera a solucao inicial.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
    	int i, j, k, p, q, m, n;

	// Inicializacao do vetor v.
	for (i = 0; i < 3; i++)
		v[i] = i + 1;
	for (i = 3; i < N; i++)
        	v[i] = 0;

	k = 3;
	while ((v[k] == 0) && (k < N))
    	{
        	p = 0;
	        m = INF;
		// Percorre todas as cidades.
        	for (i = 0; i < N; i++)
        	{
			// Verifica se a cidade ja esta no vetor.
            		if (!(verifica (v, i + 1)))
            		{
		                j = 1;
				// Percorre o vetor para encontrar qual cidade gera a insercao mais barata.
                		while ((v[j] != 0) && (j < N))
                		{
					n = d[v[j-1] -1][i] + d[i][v[j] -1] - d[v[j-1] -1][v[j] -1];
					if (m > n)
                    			{
			                        m = n;
                        			p = i + 1;
			                        q = j;
                    			}
					j++;
                		}
		                n = d[v[j-1] -1][i] + d[i][v[0] -1] - d[v[j-1] -1][v[0] -1];
                		if (m > n)
                		{
					m = n;
					p = i + 1;
                    			q = j;
                		}
            		}
        	}
		// Verifica se foi encontrada a cidade que gera a insercao mais barata.
	        if (p)
			// Insere a cidade no vetor, na posicao da insercao mais barata.
        		insere (v, p, q);
	        k++;
    	}
}

int main ()
{
	char c[T], tipo;
	int i, v[N], **d;
	float x[N], y[N];
	
	// Alocacao de memoria para a matriz de distancias.
	d = (int **) malloc (N * sizeof (int *));
	if (!d)
	{
		printf("Memoria Insuficiente!\n"); // Mensagem de erro.
		exit (0);
	}
	for(i=0; i<N; i++)
	{
		d[i] = (int *) malloc (N * sizeof (int));
		if (!d[i])
		{
			printf("Memoria Insuficiente!\n"); // Mensagem de erro.
			exit (0);
		}
	}
	
	// Leitura do arquivo de entrada e do tipo do arquivo.
	printf ("\n\nDigite o nome do arquivo de entrada: "); // <nome do arquivo>.txt
	scanf ("%s", c);
	fflush (stdin);
	printf ("\n\nInforme o tipo do arquivo (ATT = a, EUC = e, GEO = g, Matriz = m): ");
	scanf (" %c", &tipo);

	// Verificacao do tipo do arquivo e calculo da distancia.
	switch (tipo)
	{
		case 'a': le_arq (x, y, c);
			  att (x, y, d);
			  break;
		case 'e': le_arq (x, y, c);
			  euc (x, y, d);
			  break;
		case 'g': le_arq (x, y, c);
			  geo (x, y, d);
			  break;
		case 'm': le_arq_matriz (d, c);
	}

	// Obtencao da solucao inicial.
	insercao (v, d);	

	// Imprime a solucao inicial.
	for (i = 0; i < N; i++)
		printf ("\nv[%d] = %d", i, v[i]);

	return 0;
}
